# -*- coding: utf-8 -*-

from django import forms
from django.urls import reverse
from django.core import validators

from .widgets import FineInput


class FineFileField(forms.FileField):
    formid_field_name = 'formid'
    file_limit, size_limit = 4, 10485760  # 4 files by 10Mb
    widget = FineInput

    # TODO: configurable endpoints

    def __init__(self, *args, **kwargs):
        if 'file_limit' in kwargs:
            self.file_limit = kwargs.pop('file_limit')
        if 'size_limit' in kwargs:
            self.file_limit = kwargs.pop('size_limit')
        super(FineFileField, self).__init__(*args, **kwargs)

        self.widget.formid_field_name = self.formid_field_name
        self.widget.file_limit = self.file_limit
        self.widget.size_limit = self.size_limit

    def widget_attrs(self, widget):
        attrs = super(FineFileField, self).widget_attrs(widget)

        attrs['multiple'] = 'multiple'
        return attrs

    def to_python(self, data):
        if data in validators.EMPTY_VALUES:
            return None
        elif isinstance(data, list):
            return [
                super(FineFileField, self).to_python(f) for f in data
            ]
        else:
            return [data]

    def bound_data(self, data, initial):
        result = []

        if initial:
            result += initial if isinstance(initial, list) else [initial]

        if data:
            result += data if isinstance(data, list) else [data]

        return result



# -*- coding: utf-8 -*-

import os
import inspect
from datetime import timedelta

from django.db import models
from django.utils import timezone
from django.core.urlresolvers import get_callable

from django.core.files import File
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from .managers import TemporaryManager
from .utils import slugify
from .conf import settings


def upload_path(instance, filename):
    FILENAME_FUNCTION = getattr(
        settings, 'FINEUPLOADER_FILENAME_FUNCTION', None)

    func = FILENAME_FUNCTION
    if func is None:
        func = lambda x: x

    if isinstance(func, str):
        func = get_callable(func)

    return os.path.join(
        slugify(instance.__name__) if inspect.isclass(instance) else slugify(instance.__class__.__name__),
        func(filename),
    )


@python_2_unicode_compatible
class Temporary(models.Model):

    formid = models.CharField(max_length=128)
    field_name = models.CharField(max_length=256, null=True, blank=True)

    file_obj = models.FileField(_("file"), max_length=255, upload_to=upload_path)
    original_filename = models.CharField(_("original filename"), max_length=255, blank=True, null=True)

    # for internal use...

    uuid = models.UUIDField()

    timestamp = models.DateTimeField(default=timezone.now)

    objects = TemporaryManager()

    class Meta(object):
        verbose_name = _('temporary upload')
        verbose_name_plural = _('temporary uploads')
        # Query string to get back existing uploaded file is using form_id and field_name
        index_together = (
            ('formid', 'field_name'),
        )
        ordering = ['-timestamp']

    def __str__(self):
        if self.original_filename not in ('', None):
            return self.original_filename
        else:
            return self.file_obj.name

    @property
    def is_expired(self):
        EXPIRY_AGE = settings.FINEUPLOADER_TEMPORARY_AGE

        if (self.created + timedelta(seconds=EXPIRY_AGE)) <= timezone.localtime(timezone.now()):
            return True
        else:
            return False

    def as_file(self):
        class TemporaryFile(File):
            uuid = str(self.uuid)

        return TemporaryFile(self.file_obj, self.original_filename)

    def delete(self, *args, **kwargs):
        if self.file_obj and self.file_obj.storage.exists(self.file_obj.name):
            self.file_obj.delete()

        super(Temporary, self).delete(*args, **kwargs)



# -*- coding: utf-8 -*-

import json

from django.forms import ClearableFileInput

from .models import Temporary


class FineInput(ClearableFileInput):
    template_name = 'fineuploader/fineinput.html'

    def __init__(self, file_limit=4, size_limit=10485760, attrs=None):
        self.file_limit, self.size_limit = file_limit, size_limit
        super(FineInput, self).__init__(attrs)

    def render(self, name, value, attrs=None, renderer=None):
        formid_field_name = self.formid_field_name

        context = self.get_context(name, value, attrs)
        context['widget'].update({
            'formid_field_name': formid_field_name,
            'options': {
                'fileLimit': self.file_limit,
                'sizeLimit': self.size_limit,
            },
        })
        return self._render(self.template_name, context, renderer)

    def value_from_datadict(self, data, files, name):
        files = super(FineInput, self).value_from_datadict(data, files, name)
        if bool(files) is True:  # suddenly files!
            return files

        formid = data.get(self.formid_field_name, None)
        if bool(formid) is True:
            return [
                t.as_file() for t in Temporary.objects.filter(
                    formid=formid, field_name=name)
            ]
        return None



# -*- coding: utf-8 -*-

import uuid

from django import forms
from django.utils import six

from .models import Temporary
from .formfields import FineFileField


class FineFormMixin(object):
    formid_field_name = 'formid'
    request = None

    def __init__(self, *args, **kwargs):
        if kwargs.get('initial', None) is None:
            kwargs['initial'] = {}
        if not self.formid_field_name in kwargs['initial']:
            kwargs['initial'].update({
                self.formid_field_name: str(uuid.uuid4())})
        super(FineFormMixin, self).__init__(*args, **kwargs)

        self.fields[self.formid_field_name] = forms.CharField(
            widget=forms.HiddenInput, initial=kwargs['initial'][self.formid_field_name], required=False)

        formid = self.data.get(self.formid_field_name, self.initial.get(self.formid_field_name))
        for f in self.fields:
            if not isinstance(self.fields[f], FineFileField):
                continue
            
            self.fields[f].formid_field_name = self.formid_field_name

    def handle_uploads(self, *args, **kwargs):
        for f in self.fields:
            if not isinstance(self.fields[f], FineFileField):
                continue

            for file_obj in self.cleaned_data[f]:
                self.save_file(file_obj, f, *args, **kwargs)

    def save_file(self, file_obj, field_name, *args, **kwargs):
        raise NotImplementedError



# -*- coding: utf-8 -*-

import logging

from django.utils import timezone
from django.core.files.base import ContentFile
from django.core.urlresolvers import get_callable

from .ajaxuploader.backends import local as backend
from .models import Temporary
from .conf import settings

logger = logging.getLogger(__name__)


class LocalUploadBackend(backend.LocalUploadBackend):
    FILENAME_FUNCTION = getattr(
        settings, 'FINEUPLOADER_FILENAME_FUNCTION', None)

    def setup(self, request, filename, *args, **kwargs):
        super(LocalUploadBackend, self).setup(
            request, filename, *args, **kwargs)
        self.request = request  # save request for a later use

    def upload_complete(self, request, filename, *args, **kwargs):
        try:
            response = super(LocalUploadBackend, self).upload_complete(
                request, filename, *args, **kwargs)
        except Exception, e:
            return self.failure(unicode(e), request)

        return response

    def failure(self, msg, request):
        self.log_exception(msg, request)
        return {'success': False, 'error': msg, 'preventRetry': True}

    def update_filename(self, request, filename, *args, **kwargs):
        # use custom photo filename function if any
        func = self.FILENAME_FUNCTION
        if func is None:
            func = lambda x: x

        # find callable by string
        if isinstance(func, str):
            func = get_callable(func)

        value = super(LocalUploadBackend, self).update_filename(
            request, func(filename), *args, **kwargs)
        return value

    def upload(self, uploaded, filename, raw_data, *args, **kwargs):
        success = super(LocalUploadBackend, self).upload(
            uploaded, filename, raw_data, *args, **kwargs)
        if success is False:
            self.log_exception("Failed to upload: %s" % filename, self.request)
        return success

    def log_exception(self, msg, request):
        logger.error(msg, exc_info=True, extra={'stack': True})


class FineUploadBackend(LocalUploadBackend):

    def upload_complete(self, request, filename, *args, **kwargs):
        response = super(FineUploadBackend, self).upload_complete(
            request, filename, *args, **kwargs)

        if request.POST.get('qqfilename'):
            original_filename = request.POST['qqfilename']
        else:
            original_filename = request.FILES['qqfile'].name

        model_info = {
            'uuid': request.POST['qquuid'],
            'formid': request.POST['formid'],
            'original_filename': original_filename,
            'timestamp': timezone.localtime(timezone.now()),
        }

        field_name = request.POST.get('field_name')
        if field_name:
            model_info['field_name'] = field_name

        t = Temporary(**model_info)

        with open(self._path) as fh:
            t.file_obj.save(filename, ContentFile(fh.read()), save=True)

        t.save()

        return response



# -*- coding: utf-8 -*-

try:
    import json
except ImportError:
    from django.utils import simplejson as json
from django.core.serializers.json import DjangoJSONEncoder
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse, Http404, HttpResponseNotAllowed

from .ajaxuploader.views import AjaxFileUploader
from .backends import FineUploadBackend
from .models import Temporary

request_endpoint = AjaxFileUploader(backend=FineUploadBackend)


# TODO: replace with class-based view
def session_endpoint(request, *args, **kwargs):
    if request.method == "GET":
        params = {
            'formid': request.GET['formid'],
        }

        if request.GET.get('field_name'):
            params['field_name'] = request.GET.get('field_name')
        
        response = []
        for t in Temporary.objects.filter(**params).order_by('timestamp'):
            response.append({
                'name': unicode(t),
                'uuid': str(t.uuid),
                'size': t.file_obj.size,
            })

        # although "application/json" is the correct content type, IE throws a fit
        return HttpResponse(json.dumps(response, cls=DjangoJSONEncoder), content_type="text/html; charset=utf-8")
    else:
        return HttpResponseNotAllowed(['GET'])


# TODO: replace with class-based view
def delete_endpoint(request, *args, **kwargs):
    if request.method == "POST":
        try:
            Temporary.objects.get(uuid=request.POST['qquuid']).delete()
        except ObjectDoesNotExist, e:
            raise Http404

        # although "application/json" is the correct content type, IE throws a fit
        return HttpResponse(json.dumps({'success': True}, cls=DjangoJSONEncoder), content_type="text/html; charset=utf-8")
    else:
        return HttpResponseNotAllowed(['POST'])



# -*- coding: utf-8 -*-

import uuid
import pathlib

from django.utils.text import slugify as default_slugify

slugify = lambda x: default_slugify(x).lower().strip()

def parse_filename(filename):
    path = pathlib.Path(filename)

    return path.stem, path.suffix.lower()

def get_valid_filename(filename):
    stem, suffix = parse_filename(filename)

    return "{uuid}{suffix}".format(uuid=uuid.uuid4(), suffix=suffix)



# -*- coding: utf-8 -*-

from django.conf import settings  # pylint: disable=W0611

from appconf import AppConf


class FineUploaderSettings(AppConf):
    FILENAME_FUNCTION = 'fineuploader.utils.get_valid_filename'
    TEMPORARY_AGE = 60 * 60 * 24 * 7  # default is one week

    class Meta:
        prefix = 'fineuploader'



# -*- coding: utf-8 -*-

import os

from io import FileIO, BufferedWriter

from django.conf import settings

from .base import AbstractUploadBackend


class LocalUploadBackend(AbstractUploadBackend):
    UPLOAD_DIR = getattr(settings, "UPLOAD_DIR", "uploads")

    def get_path(self, filename, *args, **kwargs):
        return os.path.join(settings.MEDIA_ROOT, self.UPLOAD_DIR, filename)

    def setup(self, request, filename, *args, **kwargs):
        self._path = self.get_path(filename, *args, **kwargs)
        try:
            os.makedirs(os.path.realpath(os.path.dirname(self._path)))
        except OSError:
            pass
        self._dest = BufferedWriter(FileIO(self._path, "w"))

    def upload_chunk(self, chunk, *args, **kwargs):
        self._dest.write(chunk)

    def upload_complete(self, request, filename, *args, **kwargs):
        path = settings.MEDIA_URL + self.UPLOAD_DIR + "/" + filename
        self._dest.close()
        return {"path": path}

    def update_filename(self, request, filename, *args, **kwargs):
        """
        Returns a new name for the file being uploaded.
        Ensure file with name doesn't exist, and if it does,
        create a unique filename to avoid overwriting
        """
        self._dir = os.path.dirname(self.get_path(filename, *args, **kwargs))
        unique_filename = False
        filename_suffix = 0

        # Check if file at filename exists
        if os.path.isfile(os.path.join(self._dir, filename)):
            while not unique_filename:
                try:
                    if filename_suffix == 0:
                        open(os.path.join(self._dir, filename))
                    else:
                        filename_no_extension, extension = os.path.splitext(filename)
                        open(os.path.join(self._dir, filename_no_extension + str(filename_suffix) + extension))
                    filename_suffix += 1
                except IOError:
                    unique_filename = True

        if filename_suffix == 0:
            return filename
        else:
            return filename_no_extension + str(filename_suffix) + extension



# -*- coding: utf-8 -*-

from django.conf.urls import url

from . import views

urlpatterns = [
    url(r'^request/$', views.request_endpoint, name="request_endpoint"),
    url(r'^session/$', views.session_endpoint, name="session_endpoint"),
    url(r'^delete/$', views.delete_endpoint, name="delete_endpoint"),
]



# -*- coding: utf-8 -*-

from django.apps import AppConfig as DefaultAppConfig


class AppConfig(DefaultAppConfig):
    name = 'fineuploader'

    def ready(self):
        # Ensure everything below is only ever run once
        if getattr(AppConfig, 'has_run_ready', False):
            return
        AppConfig.has_run_ready = True



# -*- coding: utf-8 -*-

from django.contrib import admin

from .models import Temporary


@admin.register(Temporary)
class TemporaryAdmin(admin.ModelAdmin):
    list_display = ['original_filename', 'timestamp']
    readonly_fields = ['formid', 'field_name', 'timestamp']
    date_hierarchy = 'timestamp'



# -*- coding: utf-8 -*-
# Generated by Django 1.11.11 on 2018-04-05 09:55
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
import fineuploader.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Temporary',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('formid', models.CharField(max_length=128)),
                ('field_name', models.CharField(blank=True, max_length=256, null=True)),
                ('file_obj', models.FileField(max_length=255, upload_to='uploads/', verbose_name='file')),
                ('original_filename', models.CharField(blank=True, max_length=255, null=True, verbose_name='original filename')),
                ('uuid', models.UUIDField()),
                ('timestamp', models.DateTimeField(default=django.utils.timezone.now)),
            ],
            options={
                'ordering': ['-timestamp'],
                'verbose_name': 'temporary upload',
                'verbose_name_plural': 'temporary uploads',
            },
        ),
        migrations.AlterIndexTogether(
            name='temporary',
            index_together=set([('formid', 'field_name')]),
        ),
    ]
